#ifndef CHARACTER_H
#define CHARACTER_H

#include <string>
#include <vector>
#include "objects.h"
#include "../common/sources/Map.h"
#include "timer.h"
#include "attacks.h"

#define BACKPACK_SIZE   17

#define MANA_REGEN_TIMER        (30 * 60)
#define MANA_REGEN_TIMER_SLEEP  (4 * 60)
#define HEALTH_REGEN_TIMER        (30 * 60)
#define HEALTH_REGEN_TIMER_SLEEP  90
#define FOOD_TIMER                404
#define FOOD_TIMER_SLEEP          40
#define WATER_TIMER               808
#define WATER_TIMER_SLEEP         81

#define MIN_FOOD    -1024
#define MAX_FOOD    2048

class CCharacter
{
public:
    enum EBodyParts
    {
        eBodyPartHead = 0,
        eBodyPartNeck,
        eBodyPartTorso,
        eBodyPartLeftHand,
        eBodyPartRightHand,
        eBodyPartLegs,
        eBodyPartFeet,
        eBodyPartPouch1,
        eBodyPartPouch2,
        eBodyPartQuiver1,
        eBodyPartQuiver2,
        eBodyPartQuiver3,
        eBodyPartQuiver4,
        eBodyPartBackpack,
        eBodyPartCount = eBodyPartBackpack + BACKPACK_SIZE
    };

    enum EStatsNames
    {
        eStatHealth = 0,
        eStatStamina,
        eStatMana,
        eStatStrength,
        eStatDexterity,
        eStatWisdom,
        eStatVitality,
        eStatAntiMagic,
        eStatAntiFire,
        eStatLoad,
        eStatCount
    };

    enum ESkillsNames
    {
        eSkillFighter = 0,
        eSkillNinja,
        eSkillPriest,
        eSkillWizard,
        eSkillSwing,
        eSkillThrust,
        eSkillClub,
        eSkillParry,
        eSkillSteal,
        eSkillFight,
        eSkillThrow,
        eSkillShoot,
        eSkillIdentify,
        eSkillHeal,
        eSkillInfluence,
        eSkillDefend,
        eSkillFire,
        eSkillAir,
        eSkillEarth,
        eSkillWater,
        eSkillCount
    };

    struct SStat
    {
        int value;
        int maxValue;
        int startValue;
    };

    struct SSkill
    {
        int experience;
        int startXP;
    };

    CCharacter();
    bool    isDead();
    void    fromDB(int num);
    void    updateLoad();
    int     overload();
    void    load(FILE* handle);
    void    save(FILE* handle);
    void    update();
    void    enterSleep();
    void    exitSleep();
    int     getDexterity();
    int     getStrength(CObject* obj);
    int     getMeleeDamage(CVec2 monsterPos, CAttacks::SAttack& attack);
    void    hitChampion(int num, int damages);
    int     getSkillLevel(ESkillsNames skill);

    static void    readCharactersDB();

    int         portrait;
    std::string firstName;
    std::string lastName;
    SStat       stats[eStatCount];
    SSkill      skills[eSkillCount];
    CObject     bodyObjects[eBodyPartCount];
    int         food;
    int         water;
    char        spell[5];
    int         displayedDamages;   // damages displayed over character's name
    CTimer      damagesTimer;

private:
    struct SDBObject
    {
        int pos;
        int type;
    };

    struct SChampDBData
    {
        std::string firstName;
        std::string lastName;
        int     portrait;
        int     stats[eStatCount];
        int     skills[eSkillCount];
        std::vector<SDBObject>  objects;
    };

    static std::vector<SChampDBData>   dataBase;
    CTimer  manaRegenTimer;
    CTimer  healthRegenTimer;
    CTimer  foodTimer;
    CTimer  waterTimer;
};

#endif // CHARACTER_H
