#include "explosions.h"
#include "sound.h"
#include "game.h"
#include "walls.h"
#include "interface.h"
#include "player.h"
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"

CExplosions  explosions;

//---------------------------------------------------------------------------------------------
CExplosions::CExplosions()
{

}

//---------------------------------------------------------------------------------------------
void CExplosions::add(EExplosions type, CVec2 mapPos, EMonsterPos monsterPos, int duration)
{
    SExplosion  explo;
    explo.type = type;
    explo.mapPos = mapPos;
    explo.monsterPos = monsterPos;
    explo.timer.set(duration, true);
    explo.startTime = duration;

    mExploList.push_back(explo);

    if (type == eExplo_Resurrect)
        sound->play(mapPos, "sound/Explosion.wav");
}

//---------------------------------------------------------------------------------------------
void CExplosions::update()
{
    std::vector<SExplosion>::iterator   it = mExploList.begin();

    while(it != mExploList.end())
    {
        if (it->timer.update() == true)
            it = mExploList.erase(it);
        else
            it++;
    }
}

//---------------------------------------------------------------------------------------------
void CExplosions::drawExplosion(QImage* image, CVec2 mapPos, CVec2 tablePos)
{
    for (size_t i = 0; i < mExploList.size(); ++i)
    {
        SExplosion& explo = mExploList[i];

        if (explo.mapPos == mapPos)
        {
            QImage  exploGfx;
            CVec2   pos;
            float   scale = 0.0;

            switch (explo.type)
            {
                case eExplo_Resurrect:
                    exploGfx = fileCache.getImage("gfx/3DView/explosions/Fireball.png");
                    pos = CVec2(57, 66);
                    scale = 110.0 / 145.0;
                    break;

                default:
                    break;
            }

            if (scale == 0.0)
                continue;

            const SWallInfos*   tabData = &gWallsInfos[tablePos.y][tablePos.x][eWallSideUp];
            if (tabData->size.x != 0)
            {
                const SWallInfos*   refWall = NULL;
                refWall = &gWallsInfos[3][2][eWallSideUp];


                // calculate the position and scale factor and draw the explosion in a temporary image
                scale *= (float)tabData->size.y / 111.0f;

                QSize   scaledSize = exploGfx.size() * scale;
                pos = pos - refWall->pos;
                pos.x = pos.x * tabData->size.x / refWall->size.x;
                pos.y = pos.y * tabData->size.y / refWall->size.y;

//                if (flip == true)
//                    pos.x = tabData->size.x - pos.x - scaledSize.width();

                pos += tabData->pos;

                QImage  tempImage(scaledSize, QImage::Format_ARGB32);
                tempImage.fill(TRANSPARENT);
                graph2D.drawImageScaled(&tempImage, CVec2(), exploGfx, scale);

                // darken the explosion based on it's distance
                static const float shadowLevels[] = {0.0f, 0.2f, 0.4f};
                float shadow = shadowLevels[WALL_TABLE_HEIGHT - 1 - tablePos.y];

                graph2D.darken(&tempImage, shadow);

                // draw the explosion on the screen.
                graph2D.drawImage(image, pos, tempImage, 0, false, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));
            }
        }
    }
}

//---------------------------------------------------------------------------------------------
// returns the row of an explosion relatively to the player's direction. 0=back, 1=center, 2=front
uint8_t CExplosions::getRow(SExplosion& explo)
{
    static const uint8_t monsterRow[4][5] =
    {
        // NW NE SW SE C
        {  0, 0, 2, 2, 1},   // up
        {  0, 2, 0, 2, 1},   // left
        {  2, 2, 0, 0, 1},   // down
        {  2, 0, 2, 0, 1}    // right
    };

    return monsterRow[player.dir][explo.monsterPos - eMonsterPosNorthWest];
}

//---------------------------------------------------------------------------------------------
void CExplosions::drawCloud(QImage* image, CVec2 mapPos, CVec2 tablePos, int row)
{
    for (size_t i = 0; i < mExploList.size(); ++i)
    {
        SExplosion& explo = mExploList[i];

        if (explo.mapPos == mapPos && getRow(explo) == row)
        {
            QImage  exploGfx;
            float   height = 0.0;
            float   scale = 0.0;
            bool    flip = false;
            QColor  color = TRANSPARENT;

            switch (explo.type)
            {
                case eExplo_MonsterDie:
                    exploGfx = fileCache.getImage("gfx/3DView/explosions/Poison.png");
                    height = 0.636;

                    if (explo.timer.get() > explo.startTime * 2 / 3)
                    {
                        scale = 0.374;
                    }
                    else if (explo.timer.get() > explo.startTime / 3)
                    {
                        scale = 0.237;
                        flip = true;
                    }
                    else
                    {
                        scale = 0.173;
                    }

                    color = MONSTER_CLOUD_COLOR;
                    break;

                default:
                    break;
            }

            if (scale == 0.0)
                continue;

		    // bottom of the walls (ground level)
		    static const int aGround[] = {98, 106, 125, 152};
		
		    // top of the walls (ceiling level)
		    static const int aCeiling[] = {61, 58, 52, 42};

            // calculate the position and scale factor and draw the explosion in a temporary image
            CVec2   pos = monsters.getMonsterPos2(tablePos, explo.monsterPos);

            //   height intepolation
            float   rearHeight  = height * (float)aCeiling[tablePos.y]     + (1.0 - height) * (float)aGround[tablePos.y];
            float   frontHeight = height * (float)aCeiling[tablePos.y + 1] + (1.0 - height) * (float)aGround[tablePos.y + 1];

            //   depth interpolation
            float   distance = (float)(row + 1) * 0.25;
            float   meanHeight = distance * rearHeight + (1.0 - distance) * frontHeight;

            //   rounding the result
            pos.y = (int)(meanHeight + 0.5);

            // compute the scale based on the reference position (the nearest)
            CVec2   pos0, pos1;
            float tablePosY = tablePos.y * 2 + monsterPosDir[player.dir][explo.monsterPos - eMonsterPosNorthWest][1];

            pos0 = CMonsters::getMonsterPos(WALL_TABLE_WIDTH, (WALL_TABLE_HEIGHT - 1) * 2 + 1);
            pos1 = CMonsters::getMonsterPos(WALL_TABLE_WIDTH, tablePosY);
            scale *= 2.68 * (float)(pos1.x - 112) / (float)(pos0.x - 112);

            // set color
            QImage  tempImage;
            QImage* srcImage;

            if (color.alpha() == 0)
            {
                srcImage = &exploGfx;
            }
            else
            {
                tempImage = exploGfx;
                graph2D.drawImageAtlas(&tempImage, CVec2(), exploGfx, CRect(CVec2(), CVec2(exploGfx.width(), exploGfx.height())), color);
                srcImage = &tempImage;
            }

            // scale the explosion
            QImage  scaledExplo(exploGfx.size() * scale, QImage::Format_ARGB32);
            scaledExplo.fill(TRANSPARENT);
            graph2D.drawImageScaled(&scaledExplo, CVec2(), *srcImage, scale);

            // darken the explosion based on it's distance
            float shadow = ((WALL_TABLE_HEIGHT - 1) * 2 - tablePosY - 1) * 0.13f;
            graph2D.darken(&scaledExplo, shadow);

            // draw the explosion (pos relative to its encter)
            pos.x -= scaledExplo.width() / 2;
            pos.y -= scaledExplo.height() / 2;
            graph2D.drawImage(image, pos, scaledExplo, 0, flip, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));
        }
    }
}
