#include "monsters.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "player.h"
#include "walls.h"
#include "interface.h"
#include "game.h"
#include "sound.h"
#include "explosions.h"
#include "tiles.h"
#include "doors.h"

CMonsters    monsters;

//---------------------------------------------------------------------------------------------
CMonsters::CMonsters()
{

}

//---------------------------------------------------------------------------------------------
void CMonsters::readMonstersDB()
{
    QDomDocument doc;

    QFile f("databases/monsters.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomNode    monsterNode = root.firstChild();

    while(!monsterNode.isNull())
    {
        if (monsterNode.isElement())
        {
            QDomElement monster = monsterNode.toElement();
            std::vector<CParamType> paramsTypesList;

            if (monster.tagName() == "monster")
            {
                QDomElement monsterInfo = monster.firstChild().toElement();
                SMonsterData    monsterData;

                monsterData.sideAttack = false;
                monsterData.movementTime = -1;

                while(!monsterInfo.isNull())
                {
                    if (monsterInfo.tagName() == "img_front")
                    {
                        monsterData.frontImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_side")
                    {
                        monsterData.sideImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_back")
                    {
                        monsterData.backImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_attack")
                    {
                        monsterData.attackImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "time_between_atk")
                    {
                        monsterData.timeBetweenAttack = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_display_time")
                    {
                        monsterData.attackDisplayTime = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_sound")
                    {
                        monsterData.attackSound = QString("sound/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "move_sound")
                    {
                        monsterData.moveSound = QString("sound/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "movement_time")
                    {
                        monsterData.movementTime = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "defense")
                    {
                        monsterData.defense = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "health")
                    {
                        monsterData.baseHealth = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "attack")
                    {
                        monsterData.attack = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "dexterity")
                    {
                        monsterData.dexterity = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "side_attack")
                    {
                        monsterData.sideAttack = true;
                    }
                    else if (monsterInfo.tagName() == "sight_range")
                    {
                        monsterData.sightRange = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "param")
                    {
                        QString paramType = monsterInfo.attribute("type");

                        CParamType  newParamType;

                        if (paramType == "int")
                        {
                            newParamType.mType = eParamInt;
                            newParamType.mName = monsterInfo.text();
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "enum")
                        {
                            newParamType.mType = eParamEnum;
                            newParamType.mName = monsterInfo.text();
                            newParamType.mValues = monsterInfo.attribute("values").split(";");
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "stack")
                        {
                            newParamType.mType = eParamStack;
                            paramsTypesList.push_back(newParamType);
                        }
                    }

                    monsterInfo = monsterInfo.nextSibling().toElement();
                }

                CMap::mMonstersParams.push_back(paramsTypesList);
                monstersDatas.push_back(monsterData);
            }
        }
        monsterNode = monsterNode.nextSibling();
    }
}

//---------------------------------------------------------------------------------------------
int CMonsters::getStartingLife(int type)
{
    int baseHealth = monstersDatas[type].baseHealth;

    return baseHealth * game.levelMultiplier() + RANDOM((baseHealth / 4) + 1);
}

//---------------------------------------------------------------------------------------------
void CMonsters::init()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2  group;
        int nbMonsters = mapGroup.getIntParam("nbMonsters");

        group.startPos = mapGroup.mPos;
        group.dir = (EMonsterDir)mapGroup.getEnumParam("Dir");
        group.initTimers(mapGroup.getType());

        // add objects dropped by certain types of monsters
        CObject obj;

        if (mapGroup.getType() == MONSTER_TYPE_SCREAMER)
        {
            obj.setType(OBJECT_TYPE_SCREAMER_SLICE);

            for (int j = 0; j < nbMonsters; ++j)
            {
                int nbObj = RANDOM(3);

                for (int k = 0; k < nbObj; ++k)
                {
                    CObjectStack*   stack = map.addObjectsStack(group.startPos, eWallSideMonster);
                    stack->addObject(obj);
                }
            }
        }

        // initialize monsters basic datas
        for (int j = 0; j < 4; j++)
        {
            group.mMonsters[j].dir = group.dir;
            group.mMonsters[j].pos = eMonsterPosNone;
            group.mMonsters[j].life = 0;
        }

        if (nbMonsters == 1)
        {
            group.mMonsters[0].pos = eMonsterPosCenter;
            group.mMonsters[0].life = getStartingLife(mapGroup.getType());
        }
        else
        {
            // monster positions for a given direction
            static const EMonsterPos startingPos[4][4] =
            {
                {eMonsterPosNorthWest, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosSouthEast},   // up
                {eMonsterPosSouthEast, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosNorthWest},   // down
                {eMonsterPosNorthWest, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosSouthEast},   // left
                {eMonsterPosSouthEast, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosNorthWest}    // right
            };

            for (int j = 0; j < 4; j++)
            {
                if (j < nbMonsters)
                {
                    group.mMonsters[j].pos = startingPos[group.dir][j];
                    group.mMonsters[j].life = getStartingLife(mapGroup.getType());
                }
            }
        }

        group.enterWanderState();
        monsterGroups.push_back(group);
    }
}

//---------------------------------------------------------------------------------------------
CVec2 CMonsters::getMonsterPos(float x, float y)
{
    CVec2   pos;
    pos.x = 250 * (x - 4.5) / (8.5 - y) + 112;
    pos.y = 310 / (8.5 - 0.9 * y) + 66;
    return pos;
}

//---------------------------------------------------------------------------------------------
// monster positions depending on the player direction
const float monsterPosDir[4][5][2] =
{
    //   NW          NE          SW          SE           C
    {{0.0, 0.0}, {1.0, 0.0}, {0.0, 1.0}, {1.0, 1.0},  {0.5, 0.5}},   // up
    {{1.0, 0.0}, {1.0, 1.0}, {0.0, 0.0}, {0.0, 1.0},  {0.5, 0.5}},   // left
    {{1.0, 1.0}, {0.0, 1.0}, {1.0, 0.0}, {0.0, 0.0},  {0.5, 0.5}},   // down
    {{0.0, 1.0}, {0.0, 0.0}, {1.0, 1.0}, {1.0, 0.0},  {0.5, 0.5}}    // right
};

//---------------------------------------------------------------------------------------------
CVec2 CMonsters::getMonsterPos2(CVec2 tablePos, EMonsterPos monsterPos)
{
    // table pos
    float tablePosX = tablePos.x * 2 + monsterPosDir[player.dir][monsterPos - eMonsterPosNorthWest][0];
    float tablePosY = tablePos.y * 2 + monsterPosDir[player.dir][monsterPos - eMonsterPosNorthWest][1];

    return getMonsterPos(tablePosX, tablePosY);
}

//---------------------------------------------------------------------------------------------
void CMonsters::draw(QImage* image, CVec2 mapPos, CVec2 tablePos, int row)
{
    if (tablePos.y == WALL_TABLE_HEIGHT - 1)
        return;

    int groupIndex = map.findMonsterGroupIndex(mapPos);

    if (groupIndex != -1)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[groupIndex];
        CMonsterGroup2& group = monsterGroups[groupIndex];

        // monster dir depending on the player direction
        static const EMonsterDir monsterDirDir[4][4] =
        {

            {eMonsterDirUp,    eMonsterDirDown,  eMonsterDirLeft,  eMonsterDirRight}, // up
            {eMonsterDirRight, eMonsterDirLeft,  eMonsterDirUp,    eMonsterDirDown},  // left
            {eMonsterDirDown,  eMonsterDirUp,    eMonsterDirRight, eMonsterDirLeft},  // down
            {eMonsterDirLeft,  eMonsterDirRight, eMonsterDirDown,  eMonsterDirUp}     // right
        };

        // we draw the back row monsters before the front ones
        for (int i = 0; i < 4; ++i)
        {
            CMonster&   monster = group.mMonsters[i];

            if (monster.pos != eMonsterPosNone)
            {
                // relative direction
                EMonsterDir tableDir = monsterDirDir[player.dir][monster.dir];

                // sprite
                int type = mapGroup.getType();
                QString spriteName;
                bool    flip = false;

                if (tableDir == eMonsterDirDown)
                {
                    // facing
                    if (monster.displayAttackTimer.isRunning())
                    {
                        // attack
                        spriteName = monstersDatas[type].attackImage;
                        if (spriteName.isEmpty() == true)
                            spriteName = monstersDatas[type].frontImage;
                    }
                    else
                    {
                        // normal
                        spriteName = monstersDatas[type].frontImage;
                    }
                }
                else if (tableDir == eMonsterDirUp)
                {
                    // back
                    spriteName = monstersDatas[type].backImage;
                    if (spriteName.isEmpty() == true)
                        spriteName = monstersDatas[type].frontImage;
                }
                else
                {
                    // side
                    spriteName = monstersDatas[type].sideImage;
                    if (spriteName.isEmpty() == true)
                        spriteName = monstersDatas[type].frontImage;
                    else if (tableDir == eMonsterDirLeft)
                        flip = true;
                }

                if (monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1] == (float)row / 2.0)
                {
                    // screen pos
                    QImage  sprite = fileCache.getImage(spriteName.toStdString());
                    CVec2   pos = getMonsterPos2(tablePos, monster.pos);

                    // compute the scale based on the reference position (the nearest)
                    CVec2   pos0, pos1;
                    float   scale;
                    float tablePosY = tablePos.y * 2 + monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1];

                    pos0 = getMonsterPos(WALL_TABLE_WIDTH, (WALL_TABLE_HEIGHT - 1) * 2 + 1);
                    pos1 = getMonsterPos(WALL_TABLE_WIDTH, tablePosY);
                    scale = 2.37 * (float)(pos1.x - 112) / (float)(pos0.x - 112);

                    // scale the monster in a temporary image
                    QImage  scaledMonster(sprite.size() * scale, QImage::Format_ARGB32);
                    scaledMonster.fill(TRANSPARENT);
                    graph2D.drawImageScaled(&scaledMonster, CVec2(), sprite, scale);

                    // darken the monster based on it's distance
                    float shadow = ((WALL_TABLE_HEIGHT - 1) * 2 - tablePosY - 1) * 0.13f;
                    graph2D.darken(&scaledMonster, shadow);

                    // draw the monster (pos relative to its feets)
                    pos.x -= scaledMonster.width() / 2;
                    pos.y -= scaledMonster.height() - 9 * scale;
                    graph2D.drawImage(image, pos, scaledMonster, 0, flip, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));
                }
            }
        }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsters::update()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2& group = monsterGroups[i];

        group.update(mapGroup.mPos, mapGroup.getType());
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::update(CMonsterGroup2& group, CVec2 mapPos, uint8_t type, bool isInFront)
{
    displayAttackTimer.update();

    switch (group.state)
    {
        case eMonsterState_Wandering:
            wander();
            break;

        case eMonsterState_Fight:
            fight(mapPos, type, isInFront);
            break;

        case eMonsterState_Seek:
            seek();
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::wander()
{

}

//---------------------------------------------------------------------------------------------
void CMonster::seek()
{

}

//---------------------------------------------------------------------------------------------
void CMonster::initNextAttack(uint8_t type)
{
    int minTime = monsters.monstersDatas[type].timeBetweenAttack;
    int nextTime = minTime + RANDOM(40) - 10;

    if (minTime > 150)
        nextTime += RANDOM(80) - 20;

    nextAttackTimer.set(nextTime, true);
}

//---------------------------------------------------------------------------------------------
// choose the player this monster will attack
uint8_t CMonster::chooseTarget(CVec2 mapPos)
{
    // list the possible targets
    uint8_t possibleTargets[2];
    int nbTargets = 0;

    for (int i = 0; i < 4; ++i)
    {
        if (interface.isChampionEmpty(i) == false &&
            interface.isChampionDead(i) == false &&
            player.isChampionInFront(mapPos, i) == true)
        {
            possibleTargets[nbTargets++] = i;
        }
    }

    // choose a target
    return possibleTargets[RANDOM(nbTargets)];
}

//---------------------------------------------------------------------------------------------
// TODO: luck, experience, champ max damage, wounds, poison
void CMonster::fight(CVec2 mapPos, uint8_t type, bool isInFront)
{
    if (nextAttackTimer.update() == true)
    {
        initNextAttack(type);

        // if the monster is in the front line of the group
        if (isInFront == true)
        {
            // play attack sound
            QString soundName = monsters.monstersDatas[type].attackSound;

            if (soundName.isEmpty() == false)
                sound->play(mapPos, soundName.toLocal8Bit().constData());

            // show monster's animation
            int attackTime = monsters.monstersDatas[type].attackDisplayTime;
            displayAttackTimer.set(attackTime, true);

            // choose a target champion
            int target = chooseTarget(mapPos);

            // compute monster's and champion's dexterities
            int monsterDexterity = monsters.monstersDatas[type].dexterity + game.levelMultiplier() * 2 + RANDOM(32) - 16;
            int championDexterity = game.characters[target].getDexterity();

            // check if we hit
            if (interface.isSleeping() == true ||
                monsterDexterity > championDexterity ||
                RANDOM(4) == 0)
            {
                // compute damages
                int parry = (game.characters[target].getSkillLevel(CCharacter::eSkillParry) + 1) * 2;
                int damages = monsters.monstersDatas[type].attack + game.levelMultiplier() * 2 + RANDOM(16) - parry;

                if (damages <= 1)
                {
                    // parry
                    if (RANDOM(2))
                        return;

                    damages = RANDOM(4) + 2;
                }

                // randomize damages (quite a complicated way...)
                damages /= 2;
                damages += RANDOM(damages) + RANDOM(4);
                damages += RANDOM(damages);
                damages /= 4;
                damages += RANDOM(4) + 1;

                if (RANDOM(2))
                    damages -= RANDOM(damages / 2 + 1) - 1;

                // hit the champion
                static char soundName[256];
                sprintf(soundName, "sound/Champion_Damaged_%d.wav", RANDOM(4) + 1);
                sound->play(player.pos, soundName);
                game.characters[target].hitChampion(target, damages);
            }

            // if we are sleeping, wake up
            if (interface.isSleeping() == true)
                interface.wakeUp();
        }
    }
}

//---------------------------------------------------------------------------------------------
CMonsterGroup2::CMonsterGroup2()
{
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::initTimers(uint8_t type)
{
    initMoveTimer(type);

    for (int i = 0; i < 4; ++i)
        mMonsters[i].initNextAttack(type);
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::enterWanderState()
{
    state = eMonsterState_Wandering;
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::enterFightState(CVec2 mapPos)
{
    CVec2   dist = player.pos - mapPos;

    state = eMonsterState_Fight;

    // direction towards the player
    EMonsterDir dir;

    if (dist.x > 0)
        dir = eMonsterDirRight;
    else if (dist.x < 0)
        dir = eMonsterDirLeft;
    else if (dist.y > 0)
        dir = eMonsterDirDown;
    else
        dir = eMonsterDirUp;

    // initialize each monster of the group
    for (int i = 0; i < 4; ++i)
        if (mMonsters[i].pos != eMonsterPosNone)
            mMonsters[i].dir = dir;
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::enterSeekState()
{
    state = eMonsterState_Seek;
    lastPlayerPos = player.pos;
}

//---------------------------------------------------------------------------------------------
// returns the row of a monster depending on the player position (0 = front, 1 = center, 2 = back)
uint8_t CMonsterGroup2::getMonsterRow(CVec2 mapPos, EMonsterPos monsterPos)
{
    static const uint8_t monsterRow[4][5] =
    {
        // NW NE SW SE C
        {  0, 0, 2, 2, 1},   // up
        {  0, 2, 0, 2, 1},   // left
        {  2, 2, 0, 0, 1},   // down
        {  2, 0, 2, 0, 1}    // right
    };

    int playerPosFromMonster;

    if (player.pos.x < mapPos.x)
        playerPosFromMonster = 1; // left
    else if (player.pos.x > mapPos.x)
        playerPosFromMonster = 3; // right
    else if (player.pos.y < mapPos.y)
        playerPosFromMonster = 0; // up
    else
        playerPosFromMonster = 2; // down

    return monsterRow[playerPosFromMonster][monsterPos - eMonsterPosNorthWest];
}

//---------------------------------------------------------------------------------------------
// returns if a monster is in the front row of its group, taking into account the dead ones
bool CMonsterGroup2::isMonsterInFront(CVec2 mapPos, EMonsterPos monsterPos)
{
    bool    isFrontRowAlive = false;

    for (int i = 0; i < 4; ++i)
    {
        if (mMonsters[i].pos != eMonsterPosNone &&
            getMonsterRow(mapPos, mMonsters[i].pos) < 2)
        {
            isFrontRowAlive = true;
            break;
        }
    }

    uint8_t row = getMonsterRow(mapPos, monsterPos);
    return (isFrontRowAlive == true && row < 2) ||
           (isFrontRowAlive == false && row == 2);
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::update(CVec2 mapPos, uint8_t type)
{
    // update all the monsters of the group
    for (int i = 0; i < 4; ++i)
        if (mMonsters[i].pos != eMonsterPosNone)
        {
            bool    isInFront = isMonsterInFront(mapPos, mMonsters[i].pos);
            mMonsters[i].update(*this, mapPos, type, isInFront);
        }

    // handle group behavior for each state
    switch (state)
    {
        case eMonsterState_Wandering:
            wander(mapPos, type);
            break;

        case eMonsterState_Fight:
            fight(mapPos, type);
            break;

        case eMonsterState_Seek:
            seek(mapPos, type);
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::initMoveTimer(int type)
{
    CMonsters::SMonsterData&    data = monsters.monstersDatas[type];
    int t = MAX(10, RANDOM(40) + data.movementTime - 10);
    moveTimer.set(t, true);
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::move(CVec2 mapPos, uint8_t type)
{
    CMonsters::SMonsterData&    data = monsters.monstersDatas[type];
    CMonsterGroup*  group = map.findMonsterGroup(mapPos);
    group->mPos = getNewPos(mapPos, (EWallSide)dir);

    if (data.moveSound.isEmpty() == false)
        sound->play(group->mPos, data.moveSound.toLocal8Bit().constData());
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::wander(CVec2 mapPos, uint8_t type)
{
    // transition to seek state ?
    int dist = getVisibleDistance(mapPos, player.pos, type);

    if (dist > 0)
    {
        enterSeekState();
    }
    else
    {
        // wander
        if (moveTimer.update() == true)
        {
            initMoveTimer(type);

            // list the possible directions
            EWallSide   possibleDirs[4];
            int nbPossibleDirs = 0;

            for (int dir = eWallSideUp; dir < eWallSideMax; ++dir)
                if (canMove(mapPos, (EWallSide)dir) == true)
                {
                    possibleDirs[nbPossibleDirs++] = (EWallSide)dir;
                }

            if (nbPossibleDirs > 0)
            {
                // chose a random direction and move
                dir = (EMonsterDir)possibleDirs[RANDOM(nbPossibleDirs)];
                move(mapPos, type);
            }

            // rotate monsters randomly
            for (int i = 0; i < 4; ++i)
                mMonsters[i].dir = (EMonsterDir)RANDOM(4);
        }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::fight(CVec2 mapPos, uint8_t type)
{
    // transition to seek state ?
    if (mapPos.manhattanDistance(player.pos) != 1)
        enterSeekState();
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::seek(CVec2 mapPos, uint8_t type)
{
    // transition to fight state ?
    if (mapPos.manhattanDistance(player.pos) == 1)
    {
        enterFightState(mapPos);
    }
    else
    {
        // seek
        int dist = getVisibleDistance(mapPos, player.pos, type);

        if (dist > 0)
        {
            lastPlayerPos = player.pos;
        }
        else
        {
            // We can't see the player. If we arrived at the destination, go back to wander
            if (mapPos == lastPlayerPos)
            {
                enterWanderState();
                return;
            }
        }

        if (moveTimer.update() == true)
        {
            initMoveTimer(type);

            // choose direction
            EMonsterDir primaryDir;
            EMonsterDir secondaryDir;
            primaryDir = getDirectionToTarget(mapPos, lastPlayerPos, secondaryDir);

            if (canMove(mapPos, (EWallSide)primaryDir) == true)
            {

            }
            else if (canMove(mapPos, (EWallSide)secondaryDir) == true)
            {
                primaryDir = secondaryDir;
            }
            else if (canMove(mapPos, (EWallSide)oppositeDir(secondaryDir)) == true)
            {
                primaryDir = oppositeDir(secondaryDir);
            }
            else if (canMove(mapPos, (EWallSide)oppositeDir(primaryDir)) == true)
            {
                primaryDir = oppositeDir(primaryDir);
            }
            else
            {
                primaryDir = (EMonsterDir)4;
                enterWanderState();
            }

            // move
            if (primaryDir != 4)
            {
                dir = primaryDir;
                move(mapPos, type);

                for (int i = 0; i < 4; ++i)
                    mMonsters[i].dir = primaryDir;
            }
        }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::hitMonster(CVec2 groupPos, int monsterNum, int damages)
{
    mMonsters[monsterNum].life -= damages;
    enterFightState(groupPos);

    // dead ?
    if (mMonsters[monsterNum].life <= 0)
    {
        explosions.add(eExplo_MonsterDie, groupPos, mMonsters[monsterNum].pos, 45);
        mMonsters[monsterNum].pos = eMonsterPosNone;

        // if all monsters are dead, we remove the group
        int i;
        for (i = 0; i < 4; ++i)
        {
            if (mMonsters[i].pos != eMonsterPosNone)
                break;
        }

        if (i == 4)
        {
            // drop objects on the ground
            CObjectStack*   stack = map.findObjectsStack(startPos, eWallSideMonster);

            if (stack != NULL)
            {
                sound->play(groupPos, "sound/Wooden_Thud.wav");

                for (size_t j = 0; j < stack->getSize(); ++j)
                {
                    CObject obj = stack->getObject(j);
                    CVec2   objPos = groupPos * 2 + CVec2(RANDOM(2), RANDOM(2));
                    CObjectStack*   newStack = map.addObjectsStack(objPos, eWallSideMax);
                    newStack->addObject(obj);
                }
                map.removeObjectsStack(startPos, eWallSideMonster);
            }

            // delete group
            int groupIndex = map.findMonsterGroupIndex(groupPos);
            map.removeMonsterGroup(groupPos);
            monsters.monsterGroups.erase(monsters.monsterGroups.begin() + groupIndex);
        }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::load(FILE* handle, uint8_t type)
{
    uint8_t val8;

    fread(&val8, 1, 1, handle);
    dir = (EMonsterDir)val8;

    fread(&val8, 1, 1, handle);
    state = (EMonsterState)val8;

    fread(&startPos.x, sizeof(startPos.x), 1, handle);
    fread(&startPos.y, sizeof(startPos.y), 1, handle);

    for (int i = 0; i < 4; ++i)
        mMonsters[i].load(handle);

    initTimers(type);
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::save(FILE* handle)
{
    uint8_t val8;

    val8 = dir;
    fwrite(&val8, 1, 1, handle);

    val8 = state;
    fwrite(&val8, 1, 1, handle);

    fwrite(&startPos.x, sizeof(startPos.x), 1, handle);
    fwrite(&startPos.y, sizeof(startPos.y), 1, handle);

    for (int i = 0; i < 4; ++i)
        mMonsters[i].save(handle);
}

//---------------------------------------------------------------------------------------------
void CMonster::load(FILE* handle)
{
    uint8_t val8;

    fread(&val8, 1, 1, handle);
    pos = (EMonsterPos)val8;

    fread(&val8, 1, 1, handle);
    dir = (EMonsterDir)val8;

    fread(&life, sizeof(life), 1, handle);
}

//---------------------------------------------------------------------------------------------
void CMonster::save(FILE* handle)
{
    uint8_t val8;

    val8 = pos;
    fwrite(&val8, 1, 1, handle);

    val8 = dir;
    fwrite(&val8, 1, 1, handle);

    fwrite(&life, sizeof(life), 1, handle);
}

//---------------------------------------------------------------------------------------------
CVec2 CMonsterGroup2::getNewPos(CVec2 pos, EWallSide side)
{
    CVec2   newPos = pos;

    switch (side)
    {
    case eWallSideUp:
        newPos.y--;
        break;

    case eWallSideDown:
        newPos.y++;
        break;

    case eWallSideLeft:
        newPos.x--;
        break;

    case eWallSideRight:
        newPos.x++;
        break;

    default:
        break;
    }

    return newPos;
}

//---------------------------------------------------------------------------------------------
bool CMonsterGroup2::canMove(CVec2 pos, EWallSide side)
{
    CVec2   newPos = getNewPos(pos, side);

    CTile*  tile = map.getTile(pos);
    CTile*  destTile = map.getTile(newPos);

    // hit a wall
    if (tile->mWalls[side].getType() != 0)
        return false;

    // out of the map
    if (destTile == NULL)
        return false;

    switch (destTile->getType())
    {
        case eTileDoor:
            if (doors.isOpened(destTile) == false)
                return false;
            break;

        case eTileStairs:
        case eTileTeleporter:
            return false;

        case eTilePit:
            if (destTile->getBoolParam("isClosed") == false)
                return false;
            break;

        default:
            break;
    };

    // monster group
    if (map.findMonsterGroupIndex(newPos) != -1)
        return false;

    // player
    if (newPos == player.pos)
        return false;

    return true;
}

//---------------------------------------------------------------------------------------------
// returns distance between the group and the target or 0 if the group can't see the target
int CMonsterGroup2::getVisibleDistance(CVec2 pos, CVec2 target, uint8_t type)
{
    CMonsters::SMonsterData&    data = monsters.monstersDatas[type];

    for (int i = 0; i < 4; ++i)
    {
        if (mMonsters[i].pos != eMonsterPosNone)
        {
            if (data.sideAttack == true ||
                isTargetInConeOfVision(pos, target, mMonsters[i].dir) == true)
            {
                int range = data.sightRange;
                int light = game.getLightPower();
                range -= (256 - light) / 102;

                if (range < 1)
                    range = 1;

                int dist = pos.manhattanDistance(target);

                if (dist > range)
                    return 0;
                else
                    return distanceWithWalls(pos, target);
            }
        }
    }
    return 0;
}

//---------------------------------------------------------------------------------------------
bool CMonsterGroup2::isTargetInConeOfVision(CVec2 pos, CVec2 target, EMonsterDir dir)
{
    int distA, distB;

    if (dir == eMonsterDirUp)
    {
        distA = pos.y - target.y;
        distB = ABS(pos.x - target.x);
    }
    else if (dir == eMonsterDirDown)
    {
        distA = target.y - pos.y;
        distB = ABS(pos.x - target.x);
    }
    else if (dir == eMonsterDirLeft)
    {
        distA = pos.x - target.x;
        distB = ABS(pos.y - target.y);
    }
    else
    {
        distA = target.x - pos.x;
        distB = ABS(pos.y - target.y);
    }

    return (distA > 0 && distB <= distA);
}

//---------------------------------------------------------------------------------------------
// returns the distance between source and dest if there are no walls in between, or returns 0 if there are walls
int CMonsterGroup2::distanceWithWalls(CVec2 source, CVec2 dest)
{
    CVec2   delta = dest - source;
    CVec2   absDist = CVec2::abs(delta);

    // if we are right next to the destination, don't take the walls into account.
    if (absDist.x + absDist.y <= 1)
        return 1;

    // is the line more horizontal or more vertical ?
    bool    isXSmallerThanY = (absDist.x <  absDist.y);

    // compute the steps for advancing one tile at a time along the largest axis
    float   stepX, stepY;

    if (isXSmallerThanY == true)
    {
        stepY = SGN(delta.y);
        stepX = stepY * (float)delta.x / (float)delta.y;
    }
    else
    {
        stepX = SGN(delta.x);
        stepY = stepX * (float)delta.y / (float)delta.x;
    }

    // for symetry reasons, start at the center of the tile
    float   posX = (float)source.x + 0.5;
    float   posY = (float)source.y + 0.5;

    while (true)
    {
        // advance
        float nextX = posX + stepX;
        float nextY = posY + stepY;

        // test walls
        CVec2   srcTile((int)posX, (int)posY);
        CVec2   nextTile((int)nextX, (int)nextY);

        if (isWallBetweenPos(srcTile, nextTile) == true)
            return 0;

        // are we at the end ?
        if (nextTile.manhattanDistance(dest) <= 1)
            return absDist.x + absDist.y;

        // update position and loop back
        posX = nextX;
        posY = nextY;
    }
}

//---------------------------------------------------------------------------------------------
bool CMonsterGroup2::isWallBetweenPos(CVec2 source, CVec2 dest)
{
    CTile*  srcTile = map.getTile(source);
    CTile*  dstTile = map.getTile(dest);

    // check walls
    if (dest.x > source.x)
    {
        if (srcTile->mWalls[eWallSideRight].getType() != 0)
            return true;
    }
    else if (dest.x < source.x)
    {
        if (srcTile->mWalls[eWallSideLeft].getType() != 0)
            return true;
    }

    if (dest.y > source.y)
    {
        if (srcTile->mWalls[eWallSideDown].getType() != 0)
            return true;
    }
    else if (dest.y < source.y)
    {
        if (srcTile->mWalls[eWallSideUp].getType() != 0)
            return true;
    }

    // check stairs
    if (dstTile->getType() == eTileStairs)
        return true;

    // check doors
    if (dstTile->getType() == eTileDoor && dstTile->getBoolParam("isOpened") == false)
    {
        // we can't see through a closed door unless it's a porticullis
        if (dstTile->getIntParam("Type") != 0)
            return true;
    }

    return false;
}

//---------------------------------------------------------------------------------------------
EMonsterDir CMonsterGroup2::nextDir(EMonsterDir dir)
{
    static const EMonsterDir nextDirs[4] =
    {
        eMonsterDirRight,   // up
        eMonsterDirLeft,    // down
        eMonsterDirUp,      // left
        eMonsterDirDown     // right
    };

    return nextDirs[dir];
}

//---------------------------------------------------------------------------------------------
EMonsterDir CMonsterGroup2::prevDir(EMonsterDir dir)
{
    static const EMonsterDir prevDirs[4] =
    {
        eMonsterDirLeft,    // up
        eMonsterDirRight,   // down
        eMonsterDirDown,    // left
        eMonsterDirUp       // right
    };

    return prevDirs[dir];
}

//---------------------------------------------------------------------------------------------
EMonsterDir CMonsterGroup2::oppositeDir(EMonsterDir dir)
{
    static const EMonsterDir oppositeDirs[4] =
    {
        eMonsterDirDown,    // up
        eMonsterDirUp,      // down
        eMonsterDirRight,   // left
        eMonsterDirLeft     // right
    };

    return oppositeDirs[dir];
}

//---------------------------------------------------------------------------------------------
EMonsterDir CMonsterGroup2::getDirectionToTarget(CVec2 pos, CVec2 target, EMonsterDir& secondaryDir)
{
    // group aligned with target
    if (pos.x == target.x)
    {
        secondaryDir = (RANDOM(2) ? eMonsterDirLeft: eMonsterDirRight);

        if (target.y < pos.y)
            return eMonsterDirUp;
        else
            return eMonsterDirDown;
    }

    if (pos.y == target.y)
    {
        secondaryDir = (RANDOM(2) ? eMonsterDirUp: eMonsterDirDown);

        if (target.x < pos.x)
            return eMonsterDirLeft;
        else
            return eMonsterDirRight;
    }

    // group is not aligned with target: we find in which direction we need to turn to face it
    EMonsterDir primaryDir = eMonsterDirUp;

    while (true)
    {
        if (isTargetInConeOfVision(pos, target, primaryDir) == true)
        {
            // found a direction where we can see the target
            secondaryDir = nextDir(primaryDir);

            if (isTargetInConeOfVision(pos, target, secondaryDir) == false)
            {
                secondaryDir = prevDir(primaryDir);

                if (isTargetInConeOfVision(pos, target, secondaryDir) == false)
                {
                    // target can neither be seen when we turn right nor when we turn left, so we are in the "middle" of the cone
                    if (RANDOM(2))
                        secondaryDir = oppositeDir(secondaryDir);

                    return primaryDir;
                }
            }

            // we can see the target both in the primary and secondary direction, it is on the edge of the cone
            if (RANDOM(2))
            {
                EMonsterDir temp = primaryDir;
                primaryDir = secondaryDir;
                secondaryDir = temp;
            }

            return primaryDir;
        }

        primaryDir = (EMonsterDir)(primaryDir + 1);
    }
}
