#ifndef OBJECTS_H
#define OBJECTS_H

#include <string>
#include <vector>
#include <QImage>
#include "../common/sources/system/cmaths.h"
#include "../common/sources/Map.h"

#define OBJECT_TYPE_TORCH           3
#define OBJECT_TYPE_SCROLL          105
#define OBJECT_TYPE_BONES           149
#define OBJECT_TYPE_DIAMOND_EDGE    16
#define OBJECT_TYPE_HARDCLEAVE      20
#define OBJECT_TYPE_SCREAMER_SLICE  124
#define OBJECT_TYPE_EMPTYFLASK      106
#define OBJECT_TYPE_WATERFLASK      117
#define OBJECT_TYPE_WATERSKIN       144
#define OBJECT_TYPE_WATERSKIN_EMPTY 145

#define TORCH_MAX_CHARGES       114360
#define WATERSKIN_MAX_CHARGES   3

#define WEAPON_CLASS_SWING            0 /* Class 0: SWING weapons */
#define WEAPON_CLASS_DAGGER_AND_AXES  2 /* Class 1 to 15: THROW weapons */
#define WEAPON_CLASS_BOW_AMMUNITION   10
#define WEAPON_CLASS_SLING_AMMUNITION 11
#define WEAPON_CLASS_POISON_DART      12
#define WEAPON_CLASS_FIRST_BOW        16 /* Class 16 to 111: SHOOT weapons */
#define WEAPON_CLASS_LAST_BOW         31
#define WEAPON_CLASS_FIRST_SLING      32
#define WEAPON_CLASS_LAST_SLING       47
#define WEAPON_CLASS_FIRST_MAGIC      112 /* Class 112 to 255: Magic and special weapons */

class CObjects
{
public:
    struct CObjectInfo
    {
        std::string name;
        QString     category;
        QString     floorImage;
        QString     imageFile;
        uint8_t     imageNum;
        QString     positions;
        int         weight;
        int         combo;
        int         damages;
        uint8_t     weaponClass;
        int         food;
        int         water;
    };

    CObjects();

    void    readObjectsDB();
    void    drawBackRow(QImage* image, CVec2 mapPos, CVec2 tablePos);
    void    drawFrontRow(QImage* image, CVec2 mapPos, CVec2 tablePos);
    void    drawObjectsStack(QImage* image, CVec2 mapPos, EWallSide side, CVec2 tablePos);
    void    setDefaultParams(CObject* obj);
    int     getWeight(CObject* obj);

    std::vector<CObjectInfo>    mObjectInfos;

private:
    CVec2   getObjectPos(CVec2 tablePos);
    CVec2   getAlcovePos(CVec2 tablePos);
    bool    isWallInFront();
};

extern CObjects objects;

#endif // OBJECTS_H
